@@ -1,28 +1,3 @@
-/*
-  Racket
-  Copyright (c) 2004-2018 PLT Design Inc.
-  Copyright (c) 1995-2001 Matthew Flatt
-
-    This library is free software; you can redistribute it and/or
-    modify it under the terms of the GNU Library General Public
-    License as published by the Free Software Foundation; either
-    version 2 of the License, or (at your option) any later version.
-
-    This library is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
-    Library General Public License for more details.
-
-    You should have received a copy of the GNU Library General Public
-    License along with this library; if not, write to the Free
-    Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
-    Boston, MA 02110-1301 USA.
-
-  libscheme
-  Copyright (c) 1994 Brent Benson
-  All rights reserved.
-*/
-
 /* This file implements a bytecode pass to insert hook that trigger
    JIT compilation. This pass is performed after bytecode is marshaled
    or unmarshaled.
@@ -32,6 +7,7 @@
 
 #include "schpriv.h"
 #include "schrunst.h"
+#include "schmach.h"
 
 THREAD_LOCAL_DECL(static Scheme_Object *current_linklet_native_lambdas);
 static int force_jit;
@@ -445,14 +421,26 @@ Scheme_Object *scheme_case_lambda_jit(Scheme_Object *expr)
       ((Scheme_Lambda *)val)->name = name;
       if (((Scheme_Lambda *)val)->closure_size)
 	all_closed = 0;
-      if (current_linklet_native_lambdas)
-        current_linklet_native_lambdas = scheme_make_pair(val, current_linklet_native_lambdas);
     }
 
     /* Generating the code may cause empty closures to be formed: */
     ndata = scheme_generate_case_lambda(seqout);
     seqout->native_code = ndata;
 
+    if (current_linklet_native_lambdas) {
+      for (i = 0; i < cnt; i++) {
+        val = seqout->array[i];
+        {
+          /* Force jitprep on body, too, to discover all lambdas */
+          Scheme_Object *body;
+          body = jit_expr(((Scheme_Lambda *)val)->body);
+          ((Scheme_Lambda *)val)->body = body;
+        }
+        val = (Scheme_Object *)((Scheme_Lambda *)val)->u.native_code;
+        current_linklet_native_lambdas = scheme_make_pair(val, current_linklet_native_lambdas);
+      }
+    }
+
     if (all_closed) {
       /* Native closures do not refer back to the original bytecode,
 	 so no need to worry about clearing the reference. */
@@ -580,6 +568,13 @@ Scheme_Object *scheme_jit_closure(Scheme_Object *code, Scheme_Object *context)
 
     if (!context)
       data->u.jit_clone = data2;
+
+    if (current_linklet_native_lambdas) {
+      /* Force jitprep on body, too, to discover all lambdas */
+      Scheme_Object *body;
+      body = jit_expr(data2->body);
+      data2->body = body;
+    }
   }
 
   /* If it's zero-sized, then create closure now */
@@ -596,10 +591,33 @@ Scheme_Object *scheme_jit_closure(Scheme_Object *code, Scheme_Object *context)
 /*                            expressions                                 */
 /*========================================================================*/
 
+static Scheme_Object *jit_expr_k(void)
+{
+  Scheme_Thread *p = scheme_current_thread;
+  Scheme_Object *expr = (Scheme_Object *)p->ku.k.p1;
+
+  p->ku.k.p1 = NULL;
+
+  return jit_expr(expr);
+}
+
 static Scheme_Object *jit_expr(Scheme_Object *expr)
 {
   Scheme_Type type = SCHEME_TYPE(expr);
 
+#ifdef DO_STACK_CHECK
+  {
+# include "mzstkchk.h"
+    {
+      Scheme_Thread *p = scheme_current_thread;
+
+      p->ku.k.p1 = (void *)expr;
+
+      return scheme_handle_stack_overflow(jit_expr_k);
+    }
+  }
+#endif
+
   switch (type) {
   case scheme_application_type:
     return jit_application(expr);
@@ -667,6 +685,9 @@ Scheme_Linklet *scheme_jit_linklet(Scheme_Linklet *linklet, int step)
   Scheme_Object *bodies, *v;
   int i;
 
+  if (force_jit)
+    step = 2;
+
   if (!linklet->jit_ready) {
     new_linklet = MALLOC_ONE_TAGGED(Scheme_Linklet);
     memcpy(new_linklet, linklet, sizeof(Scheme_Linklet));
